#!/usr/bin/python3

# Rev A.01.08 - May 9, 2023
# Rev A.01.09 - May 28, 2023
#    * Changed level factor for audio gain - was way too loud on CW
# Rev A.01.10 - September 16, 2024
#   * Added S-meter code
#   * Added AGC init
#   * Added Squelch to volume calc
# The DZ Company, LLC
# This software may be distributed freely as long as credit is given to DZKit

import SoapySDR
from SoapySDR import * #SOAPY_SDR_ constants
import threading
import time
import queue
import RPi.GPIO as GPIO
import pyaudio
import numpy as np
from numpy import pi, absolute, arange
from scipy.signal import bilinear, lfilter, lfilter_zi, butter
from scipy import log
import sw_rx_globals as g
import statistics as stats
#import matplotlib.pyplot as plt
    
class Filter_coeff:
    def __init__(self, b, a):
        self.b = b
        self.a = a
  
def demod():
    # Demodulation
    #t0 = time.perf_counter()
    y = g.last_buf                          # get last buffer
    y = np.append(y, g.x)			        # append current buffer (now total length is 64512 complex elements)
    mode = g.mode
    g.dBm = -127
    if (g.smtr_enabled == True) or (mode == "NBFM"):    # Need dBm for squelch calc
        g.z = np.copy(y)                        # copy buffer into smtr buf

        # compute power
        # vpeak = sqrt (i^2 + q^2)
        # vrms = vpeak / sqrt(2)
        # vrms = sqrt((i^2 + q^2) / 2)
        # Power = Vrms^2 / 50 = (i^2 + q^2) / 100
        # dBm = 10 * log (Power/.001)
        # dBm = 10 * log (10 * (i^2 + q^2)) = 10 * log (10 * vpeak * vpeak)
        spb = len(g.z)
        z = g.z[(spb//4):(3*spb//4)]                    # always use middle of the two buffers
        vpeak = np.average(np.hypot(z.real, z.imag))	# get average of all magnitudes
        if vpeak != 0:
            g.dBm = 10 * np.log10(10 * vpeak * vpeak) - 60 # measured cal constant of -60 gets us close with RFG max, AGC off. 0 atten
     
    if mode == "FM" or mode == "NBFM":
        g.frq_ofs = 0                       # account for freq shift
        y = np.diff(np.unwrap(np.angle(y))) # demod
        b = g.fmf.b			                # get de-emphasis filter coeffs
        a = g.fmf.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
    elif mode == "AM":
        g.frq_ofs = 0                       # account for freq shift
        b = g.bpf6k.b			            # Get 6kHz filter coeffs
        a = g.bpf6k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
    elif mode == "USB":
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_m1r4k                # shift down 1.4kHz
        b = g.bpf2r8k.b			            # get +/-1400 Hz filter coeffs
        a = g.bpf2r8k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_1r4k                 # shift back up 1400Hz
    elif mode == "LSB":  
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_1r4k                 # shift up 1400Hz
        b = g.bpf2r8k.b			            # get +/-1400 Hz filter coeffs
        a = g.bpf2r8k.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_m1r4k                # shift back down 1400Hz
    elif mode == "CW(USB)":
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_m400                 # shift down 400Hz
        b = g.bpf800.b			            # get +/-400 Hz filter coeffs
        a = g.bpf800.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y = lfilter(b, a, y)        	    # apply chosen filter
        y *= g.exp_jwt_400                  # shift back up 400Hz
    elif mode == "CW(LSB)":  
        g.frq_ofs = 0                       # account for freq shift
        y *= g.exp_jwt_400                  # shift up 400Hz
        b = g.bpf800.b			            # get +/-400 Hz filter coeffs
        a = g.bpf800.a
        zi = lfilter_zi(b, a)*g.ly	   	    # init filter 
        y, zf = lfilter(b, a, y, zi=zi)	    # apply chosen filter
        y *= g.exp_jwt_m400                 # shift back down 400Hz

    g.ly = g.x[-1]                          # get last element of current buf to set initial condition of next filter
 
    if mode == "AM":
        y = np.hypot(y.real, y.imag)	    # demod 
    else:
        y = y.real                          # FM already demodulated, but just need real part of array

    y = y[::g.decimation]	                # take every 8 elements -> 8064 reals
    
    spb = len(y)
    y = y[(spb//4):(3*spb//4)]              # always use middle of the two buffers

    # make sure the gain is such that the value to be output is roughly the same per mode
    if g.mode == "FM":
        maxi = 20.0e4
    elif g.mode == "NBFM":
        maxi = 500.0e4
    elif g.mode == "AM":
        maxi = 100.0e6
    elif g.mode == "LSB" or g.mode == "USB":
        maxi = 100.0e6
    else:   # CW
        maxi = 100.0e6

    # Apply volume control
    yc = maxi * g.afg / 30000.0
    # Handle squelch on NBFM
    if g.mode == "NBFM" and g.dBm < g.squelch:
        yc = 0
    
    y *= yc
    
    # check for overload and limit the peak to 30000
    poverload = np.abs(np.max(y))/30000.0
    noverload = np.abs(np.min(y))/30000.0
    divisor = 1.0
    
    if poverload > 1.0:
        divisor = poverload
        #print ("+Overload comp ", poverload, g.iter_nr)

    if noverload > 1.0 and noverload > poverload:
        divisor = noverload
        #print ("-Overload comp ", noverload, g.iter_nr)
        
    y /= divisor
    
    # Generate max amplitude sine wave for test purposes
    #for i in range (0, 4032):
    #    y[i] = 32767 * np.sin(2*np.pi*0.032*i) * g.afg / 1024
    #print (g.afg)

    #plt.plot(y)
    #plt.show()
    
    g.audio_buf.put(y.astype(np.int16))
    
    #print("Demod done, data added to FIFO. Demod time = ", time.perf_counter()-t0)
    
    return
    
# Callback to process "framesize" chunks of the audio buffer
def audio_callback(in_data, frame_count, time_info, status):
    g.ab_frame = g.audio_buf.get()
    ret_data = g.ab_frame[0 : frame_count].astype(np.int16).tobytes()
    if g.stopped == True:
        return ret_data, pyaudio.paComplete
    return ret_data, pyaudio.paContinue   

def init_sdr (sdr):
    # SDR setup params
    bandwidth = 200e3
    rx_chan = 0 # only 1 channel on RSP1A

    # init hardware
    sdr.setSampleRate(SOAPY_SDR_RX, rx_chan, g.sample_rate)
    sdr.setFrequency(SOAPY_SDR_RX, rx_chan, g.center_freq)
    sdr.setBandwidth(SOAPY_SDR_RX, rx_chan, bandwidth)
    sdr.setGain(SOAPY_SDR_RX, rx_chan, "RFGR", g.lna)               # gain reduction values go from 0-9 or 0-5; default to 2
    sdr.writeSetting('agc_setpoint', -50)                           # assume midscale
    if g.agc == "Slow":                                             # default in global settings
        sdr.setGainMode(SOAPY_SDR_RX, rx_chan, True)                # turn on AGC
        sdr.writeSetting('agc_attack', 500)
        sdr.writeSetting('agc_decay', 3000)
        sdr.writeSetting('agc_decay_delay', 200)
        sdr.writeSetting('agc_decay_threshold', 5)	
    elif g.agc == "Fast":
        sdr.setGainMode(SOAPY_SDR_RX, rx_chan, True)                # turn on AGC
        sdr.writeSetting('agc_attack', 500)
        sdr.writeSetting('agc_decay', 500)
        sdr.writeSetting('agc_decay_delay', 200)
        sdr.writeSetting('agc_decay_threshold', 5)	
    elif g.agc == "Off":
        sdr.setGainMode(SOAPY_SDR_RX, rx_chan, False)               # turn off AGC       
        sdr.setGain(SOAPY_SDR_RX, rx_chan, "IFGR", g.rfg)           # gain reduction values go from 0-42dB; default to 20

    # Start SDR data stream
    g.rxStream = sdr.setupStream(SOAPY_SDR_RX, SOAPY_SDR_CF32, [rx_chan])
    sdr.activateStream(g.rxStream) 
    return 

def init_filts(show_chart):
    # calculate FM de-emphasis filter params
    g.fmf.b, g.fmf.a = bilinear(1, [75e-6, 1], fs=g.sample_rate)

    #calculate AM Butterworth LPF (+/-3kHz bpf including neg freq)
    wch = 2 * pi * 3000
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf6k.b, g.bpf6k.a = bilinear(b, a, fs=g.sample_rate)
    
    wch = 2 * pi * 10000
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf20k.b, g.bpf20k.a = bilinear(b, a, fs=g.sample_rate)
    
    #Shift the filter up 3KHz in frequency by multiplying by exp(j*2*pi*f0*t)
 
    # calculate Butterworth LPF for SSB (+/-2.8kHz bpf incl neg freq)
    wch = 2 * pi * 1400 
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf2r8k.b, g.bpf2r8k.a = bilinear(b, a, fs=g.sample_rate)
    
    # calculate Butterworth LPF for CW (+/-400Hz bpf incl neg freq)
    wch = 2 * pi * 400 
    b, a = butter(4, wch, btype='low', analog=True)
    g.bpf800.b, g.bpf800.a = bilinear(b, a, fs=g.sample_rate)

    # calc time interval
    Ts = 1.0/g.sample_rate                  # sample period
    t = np.arange(0.0, Ts*(g.num_samps*2), Ts)  # time vector. args are (start, stop, step) (2x buffer length)

    # Shift the filter up in frequency 3kHz by multiplying by exp(j*2*pi*f0*t)    
    f0 = 3.0e3                              # freq shift
    g.exp_jwt_3k = np.exp(2j*np.pi*f0*t)    # this is essentially a complex sine wave
    
    # Shift the filter down in frequency 1.4kHz by multiplying by exp(-j*2*pi*f0*t)
    f0 = -1.4e3
    g.exp_jwt_m1r4k = np.exp(2j*np.pi*f0*t)

    # Shift the filter up in frequency 1.4kHz by multiplying by exp(j*2*pi*f0*t)
    f0 = 1.4e3
    g.exp_jwt_1r4k = np.exp(2j*np.pi*f0*t)
    
    # Shift the filter down in frequency 400Hz by multiplying by exp(-j*2*pi*f0*t)
    f0 = -400
    g.exp_jwt_m400 = np.exp(2j*np.pi*f0*t)

    # Shift the filter up in frequency 400Hz by multiplying by exp(j*2*pi*f0*t)
    f0 = 400
    g.exp_jwt_400 = np.exp(2j*np.pi*f0*t)
    return

def init_audio():
    g.p = pyaudio.PyAudio()
    g.stream = g.p.open(format=pyaudio.paInt16,
                        channels=1,
                        rate=g.drate,
                        output=True,
                        stream_callback=audio_callback,
                        frames_per_buffer=g.framesize)
    g.stream.start_stream()
    return 
    
def read_sdr_thread(sdr):
    while (g.stopped == False):
        read_sdr(sdr)

def read_sdr (sdr):
    g.iter_nr += 1
    for i in range(g.num_rx_calls):
        sr = sdr.readStream(g.rxStream, [g.sample_buffer], len(g.sample_buffer))
        g.x[i*g.max_buffer_size:(i+1)*g.max_buffer_size] = g.sample_buffer

    demod()                                 # run demodulator and filtering algorithm on two buffers
    
    g.last_buf = np.copy(g.x)               # save current buffer for use in next double-wide calc

    if (g.iter_nr == 4): init_audio()       # start audio stream after first two buffers have been put into the fifo
    
    disp_s_mtr(g.dBm)

    return

def clear_old_rssi():
    if g.seglit == 1: g.canvas1.delete(g.seg1)
    if g.seglit == 2: g.canvas1.delete(g.seg2)
    if g.seglit == 3: g.canvas1.delete(g.seg3)
    if g.seglit == 4: g.canvas1.delete(g.seg4)
    if g.seglit == 5: g.canvas1.delete(g.seg5)
    if g.seglit == 6: g.canvas1.delete(g.seg6)
    if g.seglit == 7: g.canvas1.delete(g.seg7)
    if g.seglit == 8: g.canvas1.delete(g.seg8)
    if g.seglit == 9: g.canvas1.delete(g.seg9)
    if g.seglit == 10: g.canvas1.delete(g.seg10)
    if g.seglit == 11: g.canvas1.delete(g.seg11)
    if g.seglit == 12: g.canvas1.delete(g.seg12)
    if g.seglit == 13: g.canvas1.delete(g.seg13)
    if g.seglit == 14: g.canvas1.delete(g.seg14)
    if g.seglit == 15: g.canvas1.delete(g.seg15)
    if g.seglit == 16: g.canvas1.delete(g.seg16)
    return
                        
def disp_s_mtr(dBm):
    if g.smtr_enabled == False:
        clear_old_rssi()         
        return
    # display rssi meter
    g.smx1 = 790
    g.smy1 = 490
    g.smx2 = 951
    g.smy2 = 500

    if dBm < -121:
        if g.seglit != 1:
            clear_old_rssi() 
            g.seg1 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+11, g.smy1+9, fill="red", outline='white', width=0)  
            g.seglit = 1
    elif dBm < -115:
        if g.seglit != 2:
            clear_old_rssi() 
            g.seg2 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+21, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 2
    elif dBm < -109:
        if g.seglit != 3:
            clear_old_rssi() 
            g.seg3 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+31, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 3
    elif dBm < -97:
        if g.seglit != 4:
            clear_old_rssi() 
            g.seg4 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+41, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 4
    elif dBm < -91:
        if g.seglit != 5:
            clear_old_rssi() 
            g.seg5 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+51, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 5
    elif dBm < -85:
        if g.seglit != 6:
            clear_old_rssi()
            g.seg6 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+61, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 6
    elif dBm < -79:
        if g.seglit != 7:
            clear_old_rssi() 
            g.seg7 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+71, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 7
    elif dBm < -73:
        if g.seglit != 8:
            clear_old_rssi() 
            g.seg8 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+81, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 8
    elif dBm < -67:
        if g.seglit != 9:
            clear_old_rssi() 
            g.seg9 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+91, g.smy1+9, fill="red", outline='white', width=0)  
            g.seglit = 9
    elif dBm < -61:
        if g.seglit != 10:
            clear_old_rssi() 
            g.seg10 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+101, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 10
    elif dBm < -55:
        if g.seglit != 11:
            clear_old_rssi() 
            g.seg11 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+111, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 11
    elif dBm < -49:
        if g.seglit != 12:
            clear_old_rssi() 
            g.seg12 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+121, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 12
    elif dBm < -43:
        if g.seglit != 13:
            clear_old_rssi() 
            g.seg13 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+131, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 13
    elif dBm < -37:
        if g.seglit != 14:
            clear_old_rssi() 
            g.seg14 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+141, g.smy1+9, fill="red", outline='white', width=0)   
            g.seglit = 14
    elif dBm < -31:
        if g.seglit != 15:
            clear_old_rssi() 
            g.seg15 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+151, g.smy1+8, fill="red", outline='white', width=0)   
            g.seglit = 15
    else:
        if g.seglit != 16:
            clear_old_rssi()
            g.seg16 = g.canvas1.create_rectangle(g.smx1+1, g.smy1+1, g.smx1+161, g.smy1+8, fill="red", outline='white', width=0)  
            g.seglit = 16
    return


def shut_down_streams(sdr):         
    #shut down the SDR stream
    #print ("Deactivating sdr stream")
    sdr.deactivateStream(g.rxStream)

    #shut down the audio stream
    #print ("Stopping audio stream")
    while g.stream.is_active():
        #print ("Waiting for audio stream to go inactive")
        time.sleep(0.1)
    g.stream.stop_stream()
    #print ("Closing audio stream")
    g.stream.close()
    #print ("Terminating pyAudio")
    g.p.terminate()
    return
